@@ -61,32 +61,54 @@ bool ThreadEntrySet::basicSanity() const {
   if constexpr (!kIsDebug) {
     return true;
   }
-  if (threadEntries.empty() && entryToVectorSlot.empty()) {
+  if (threadElements.empty() && entryToVectorSlot.empty()) {
     return true;
   }
-  if (threadEntries.size() != entryToVectorSlot.size()) {
+  if (threadElements.size() != entryToVectorSlot.size()) {
     return false;
   }
-  auto const size = threadEntries.size();
+  auto const size = threadElements.size();
   rand_engine rng;
   std::uniform_int_distribution<size_t> dist{0, size - 1};
   if (dist(rng) < constexpr_log2(size)) {
     return true;
   }
   return //
+      threadElements.size() == entryToVectorSlot.size() &&
       std::all_of(
           entryToVectorSlot.begin(),
           entryToVectorSlot.end(),
           [&](auto const& kvp) {
-            return kvp.second < threadEntries.size() &&
-                threadEntries[kvp.second] == kvp.first;
+            return kvp.second < threadElements.size() &&
+                threadElements[kvp.second].threadEntry == kvp.first;
           });
 }
 
+bool ThreadEntry::cachedInSetMatchesElementsArray(uint32_t id) {
+  if constexpr (!kIsDebug) {
+    return true;
+  }
+
+  if (removed_) {
+    // Pointer in entry set and elements array need not match anymore.
+    return true;
+  }
+
+  auto rlock = meta->allId2ThreadEntrySets_[id].tryRLock();
+  if (!rlock) {
+    // Try lock failed. Skip checking in this case. Avoids
+    // getting stuck in case this validation is called when
+    // already holding the entry set lock.
+    return true;
+  }
+
+  return elements[id].ptr == rlock->getPtrForThread(this);
+}
+
 void ThreadEntrySet::compress() {
   assert(compressible());
   // compress the vector
-  threadEntries.shrink_to_fit();
+  threadElements.shrink_to_fit();
   // compress the index
   EntryIndex newIndex;
   newIndex.reserve(entryToVectorSlot.size());
@@ -144,6 +166,32 @@ bool StaticMetaBase::dying() {
   return folly::detail::thread_is_dying();
 }
 
+/*
+ * Note on safe lifecycle management of TL objects:
+ * Any instance of a TL object created has to be disposed of safely. The
+ * non-trivial cases are a) when a thread exits and we have to dispose of all
+ * instances that belong to the thread across different TL objects; b) when a TL
+ * object is destroyed and we have to dispose of all instances of that TL object
+ * across all the threads that may have instantiated it. These 2 paths can also
+ * race and are resolved, using the ThreadEntrySet locks, as follows:
+ * - The exiting thread goes over all ThreadEntrySets and removes itself from
+ * it. After this, the thread is responsible for disposing of any elements that
+ * are still set in its 'elements' array. By setting the removed_ flag any new
+ * elements created when disposing elements will not be made discoverable via
+ * ThreadEntrySets and hence remain the responsibility of the exiting thread to
+ * dispose.
+ * - The thread calling destroy goes over all ThreadEntry in the ThreadEntrySet
+ * of the TL object being destoryed (in popThreadEntrySetAndClearElementPtrs).
+ * The ElementWrapper for the ThreadEntry found there are copied out and cleared
+ * from the 'elements' array. This is done under the ThreadEntrySet lock and the
+ * 'destroy' call is responsible for disposing the elements it thus claimed. The
+ * actual dispose happens after releasing the locks and since the TL object is
+ * being destroyed, no new elements of it should be getting created. The destroy
+ * code should not access any ThreadEntry outside the safety of
+ * popThreadEntrySetAndClearElementPtrs as their owner threads could exit and
+ * release them.
+ */
+
 void StaticMetaBase::onThreadExit(void* ptr) {
   folly::detail::thread_is_dying_mark();
   auto threadEntry = static_cast<ThreadEntry*>(ptr);
@@ -164,7 +212,11 @@ void StaticMetaBase::onThreadExit(void* ptr) {
     forkRlock.unlock();
     {
       std::lock_guard g(meta.lock_);
-      // mark it as removed
+      // mark it as removed. Once this is set, any new TL object created in this
+      // thread as part of invoking the destructor on another TL object will not
+      // record this threadEntry in that TL object's ThreadEntrySet. This thread
+      // itself will be responsible for cleaning up the new object(s) so
+      // created.
       threadEntry->removed_ = true;
       auto elementsCapacity = threadEntry->getElementsCapacity();
       auto beforeCount = meta.totalElementWrappers_.fetch_sub(elementsCapacity);
@@ -215,6 +267,11 @@ void StaticMetaBase::cleanupThreadEntriesAndList(
       if (meta.strict_) {
         rlock.lock();
       }
+      // ThreadEntry 'tmp' was removed from all sets in its own cleanup. Since
+      // we set removed_ flag, no subsequent element being added (during destroy
+      // of other TL elements) should have added the ThreadEntry back to the
+      // tracking set.
+      DCHECK(meta.isThreadEntryRemovedFromAllInMap(tmp, !meta.strict_));
 
       for (bool shouldRunInner = true; shouldRunInner;) {
         shouldRunInner = false;
@@ -280,12 +337,48 @@ uint32_t StaticMetaBase::allocate(EntryID* ent) {
   return id;
 }
 
+ThreadEntrySet StaticMetaBase::popThreadEntrySetAndClearElementPtrs(
+    uint32_t id) {
+  // Lock the ThreadEntrySet for id so that no other thread can update
+  // its local ptr or alter the its elements array or ThreadEntry object
+  // itself, before this function is done updating them. This is called
+  // by the `destroy` function to release a TL object. There should be
+  // no racing accessAllThreads() call with it.
+  auto wlocked = allId2ThreadEntrySets_[id].wlock();
+  ThreadEntrySet tmp;
+  std::swap(*wlocked, tmp);
+  std::lock_guard g(lock_);
+  for (auto& e : tmp.threadElements) {
+    auto elementsCapacity = e.threadEntry->getElementsCapacity();
+    if (id < elementsCapacity) {
+      /*
+       * Writing another thread's ThreadEntry from here is fine;
+       * The TL object is being destroyed, so get(id), or reset()
+       * or accessAllThreads calls on it are illegal. Only other
+       * racing accesses would be from the owner thread itself
+       * either a) reallocating the elements array (guarded by
+       * lock_, so safe) or b) exiting and trying to clear the
+       * elements array or free the elements and ThreadEntry itself. The
+       * ThreadEntrySet lock synchronizes this part as the exiting thread will
+       * acquire it to remove itself from the set.
+       */
+      DCHECK_EQ(e.threadEntry->elements[id].ptr, e.wrapper.ptr);
+      e.wrapper.deleter = e.threadEntry->elements[id].deleter;
+      e.threadEntry->elements[id].ptr = nullptr;
+      e.threadEntry->elements[id].deleter = 0;
+    }
+    // Destroy should not access thread entry after this call as racing
+    // exit call can make it invalid.
+    e.threadEntry = nullptr;
+  }
+  return tmp;
+}
+
 void StaticMetaBase::destroy(EntryID* ent) {
   try {
     auto& meta = *this;
 
     // Elements in other threads that use this id.
-    std::vector<ElementWrapper> elements;
     ThreadEntrySet tmpEntrySet;
 
     {
@@ -307,39 +400,25 @@ void StaticMetaBase::destroy(EntryID* ent) {
       if (id == kEntryIDInvalid) {
         return;
       }
-      meta.allId2ThreadEntrySets_[id].swap(tmpEntrySet);
+      tmpEntrySet = meta.popThreadEntrySetAndClearElementPtrs(id);
       forkRlock.unlock();
 
       {
+        // Release the id to be reused by another TL variable. Some
+        // other TL object may acquire and re-use it before the rest
+        // of the cleanup work is done. That is ok. The destructor callbacks
+        // for the objects should not access the same TL variable again. If
+        // we want to tolerate that pattern, due to any legacy behavior, this
+        // block should be after the loop to dispose of all collected elements
+        // below.
         std::lock_guard g(meta.lock_);
-        for (auto& e : tmpEntrySet.threadEntries) {
-          auto elementsCapacity = e->getElementsCapacity();
-          if (id < elementsCapacity) {
-            if (e->elements[id].ptr) {
-              elements.push_back(e->elements[id]);
-              /*
-               * Writing another thread's ThreadEntry from here is fine;
-               * the only other potential reader is the owning thread --
-               * from onThreadExit (which grabs the lock, so is properly
-               * synchronized with us) or from get(), which also grabs
-               * the lock if it needs to resize the elements vector.
-               *
-               * We can't conflict with reads for a get(id), because
-               * it's illegal to call get on a thread local that's
-               * destructing.
-               */
-              e->elements[id].ptr = nullptr;
-              e->elements[id].deleter = 0;
-            }
-          }
-        }
         meta.freeIds_.push_back(id);
       }
     }
     // Delete elements outside the locks.
-    for (ElementWrapper& elem : elements) {
-      if (elem.dispose(TLPDestructionMode::ALL_THREADS)) {
-        elem.cleanup();
+    for (auto& e : tmpEntrySet.threadElements) {
+      if (e.wrapper.dispose(TLPDestructionMode::ALL_THREADS)) {
+        e.wrapper.cleanup();
       }
     }
   } catch (...) { // Just in case we get a lock error or something anyway...
@@ -478,7 +557,17 @@ FOLLY_NOINLINE void StaticMetaBase::ensureThreadEntryIsInSet(
  */
 void* ThreadEntry::releaseElement(uint32_t id) {
   auto rlocked = meta->allId2ThreadEntrySets_[id].rlock();
-  return elements[id].release();
+  auto capacity = getElementsCapacity();
+  void* ptrToReturn = (capacity >= id) ? elements[id].release() : nullptr;
+  auto slot = rlocked->getIndexFor(this);
+  if (slot < 0) {
+    DCHECK(removed_ || ptrToReturn == nullptr);
+    return ptrToReturn;
+  }
+  auto& element = rlocked.asNonConstUnsafe().threadElements[slot];
+  DCHECK_EQ(ptrToReturn, element.wrapper.ptr);
+  element.wrapper = {};
+  return ptrToReturn;
 }
 
 /*
@@ -493,6 +582,33 @@ void ThreadEntry::cleanupElement(uint32_t id) {
   elements[id].cleanup();
 }
 
+void ThreadEntry::resetElementImplAfterSet(
+    const ElementWrapper& element, uint32_t id) {
+  auto& set = meta->allId2ThreadEntrySets_[id];
+  auto rlock = set.rlock();
+  cleanupElement(id);
+  elements[id] = element;
+  if (removed_) {
+    // Elements no longer being mirrored in the ThreadEntrySet.
+    // Thread must have cleared itself from the set when it started exiting.
+    DCHECK(!rlock->contains(this));
+    return;
+  }
+  if (element.ptr != nullptr && !rlock->contains(this)) {
+    meta->ensureThreadEntryIsInSet(this, set, rlock);
+  }
+  auto slot = rlock->getIndexFor(this);
+  if (slot < 0) {
+    // Not present in ThreadEntrySet implies the value was never set to be
+    // non-null and new value in element.ptr is nullptr as well.
+    DCHECK(!element.ptr);
+    DCHECK(!elements[id].ptr);
+    return;
+  }
+  size_t uslot = static_cast<size_t>(slot);
+  rlock.asNonConstUnsafe().threadElements[uslot].wrapper = element;
+}
+
 FOLLY_STATIC_CTOR_PRIORITY_MAX
 PthreadKeyUnregister PthreadKeyUnregister::instance_;
 #if defined(__GLIBC__)
